/* Definitions for the translator from the interpreted form to SSA.
   Copyright (C) 1992-1997 Jean Goubault-Larrecq and Bull S.A.

   This file is part of HimML.

   HimML is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   HimML is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with HimML; see the file COPYING.  If not, write to
   the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/

/* we translate expressions and bindings, as submitted to
   Eval() and EvalBindings(), into a simple single-assignment
   imperative language, which we call scofl.
   	The statements are organized as a map from labels to instructions.
   	Instructions are assignments, tests, and calls.
   	(this looks like Appel's CPS, but it is not, since operations
   on continuations are done through C procedure calls).
   	The map is coded as a simple unshared list of instructions,
   and labels are just addresses of sub-lists. Variables are actually
   local ML variables (this is a single-assignment language), so we use
   labels to denote also the variable that is being assigned at this
   label, if there is one.
   	Tags are:
   	- E_CONS (assignments) the cdr is the continuation; the car
   	  is an expression expr, meaning x := expr (where x is the variable
   	  associated with the label).
   	- E_PROG is for headers of functions. This is an unshared list
	  whose cells are the input parameters themselves. Inside the
	  cells, we find either NIL (storage class unknown, assumed ExprPtr),
	  or a boxed int saying what the storage class of the variable is
	  (see optimize.h).
   	- E_IF (tests) the format is E_IF(E_CONS(x,then),else), where
   	  x is the (unboxed) boolean variable to be tested, and then and
   	  else are the two possible continuations. Notice that the idea
   	  is that continuations always lie in the CDR of some cells.
	  x may also be an E_COUNTED(tag,x) where tag may be Z_INC_LISTSWEEP,
	  Z_INC_MAPSWEEP or Z_INC_SUBMAPSWEEP, or Z_START_HANDLE.
	- E_GOTO is for merge points. The car contains a E_COUNTED.
	  The count denotes the number
	  of predecessors of this merge point, which should always be >=1.
	  The value is a map from lasts of each branch to the variable that
	  is the result of this branch.
   	- E_NIL (termination, return).
*/
/* scofl expressions are applications of primitive functions to
   previously defined variables. As there is only a finite number
   of primitive functions, we code them as integers. The right
   structure for this is an E_COUNTED cell. Its REFVAL then points
   to the argument.
*/
typedef struct compEnv {
  ExprPtr env; /* list (x1 x2 ... xn-1 . xn) of variables representing
		  the various levels in the run-time environment (xn is
		  the rest) */
  COUNTER level; /* level of the current environment (0=toplevel, ie
		    only the global environment is present) */
  ExprPtr funs; /* list of maps, modeled after real environments (except we
		   have maps from indexes to values instead of vectors),
		   to hold local definitions of functions. */
  ExprPtr handlers; /* same thing, representing exception handlers;
		       the entries are labels of exception handlers.
		       Moreover, xn is always NIL, representing the
		       handler stack at the beginning of the current function.
		     */
  ExprPtr constpatmemoref; /* only valid when compiling patterns, holds
			      map from patterns to the only value they
			      match, when they do. */
  ExprPtr letdelayref; /* only valid when compiling patterns, holds a map
			  from ranks to variables, that should then be
			  translated to Z_LET assignment expressions
			  after the pattern is compiled. */
  ExprPtr letconsenv; /* only valid when compiling patterns,
			 expression that would build the current environment,
			 when getEnv(env) is NIL */
} compEnv;

#define NTAGS ((E_UNUSED+7)>>3)

typedef struct Code { /* structure inside an E_EXTERNAL, to represent
			 lambda-expressions */
  CHAR *name; /* C name of function. */
  INT fxtypes; /* flag bits for types of possible side effects; see optimize.h */
  CHAR tags[NTAGS]; /* possible tags for output of function */
  INT size; /* used for deciding whether to inline it or not */
  INT flags;
  ExprPtr prgm; /* code, or NIL if not loaded in memory */
  ExprPtr last; /* last node for prgm (or return variable/statement) */
  MapPtr lambda; /* set of locally-defined Code functions (E_EXTERNALs) */
  INT storage; /* storage class of last; filled up in optimize.c */
  ExprPtr deco; /* ppDecoration: see optimize.h */
} Code;

/* #defines for flags field: */
#define CODE_PRIVATE 0x01
#define CODE_INLINE 0x02
#define CODE_DOESNOTUSE_ENV 0x04
#define CODE_INUSE 0x40
#define CODE_ANALYZED 0x80

/* storage classes are unions of possible real storage classes,
   represented as bits (masks have names beginning in SCF_).
   SCF_COMPOSITE means any object composed of other objects (cons cells,
   complex numbers, etc. then their description should be found in the
   value field) */
#define SCF_TOP 0
#define SCF_BOOL 0x01
#define SCF_INT 0x02
#define SCF_STRING 0x04
#define SCF_REAL0 0x08
#define SCF_REALN 0x10
#define SCF_REAL (SCF_REAL0 | SCF_REALN)
#define SCF_CODE 0x20
#define SCF_COMPOSITE 0x80
#define SCF_BASIC (SCF_BOOL | SCF_INT | SCF_STRING | SCF_REAL | SCF_CODE)
#define SCF_ALL_TYPES (SCF_BASIC | SCF_COMPOSITE)
#define SCF_LOCAL 0x200
#define SCF_BOXED 0x400
#define SCF_UNBOXED 0x800
#define SCF_ALL_BOXINGS (SCF_LOCAL | SCF_BOXED | SCF_UNBOXED)
#define SCF_NOINIT 0x8000
/* SCF_NOINIT flag is cleared if value can be computed at init time */
#define SCF_ALL (SCF_ALL_TYPES | SCF_ALL_BOXINGS | SCF_NOINIT)

typedef ExprPtr (*compileNeutralFunc) P((IN CONST compEnv *env,INOUT ExprPtr last,
					 INOUT ExprPtr *nonelastp,
					 INOUT ExprPtr *errlastp));
typedef ExprPtr (*compileCombineFunc) P((IN ExprPtr e1,
					 IN CONST compEnv *env,INOUT ExprPtr last,
					 INOUT ExprPtr *errlastp));
typedef ExprPtr (*compileCombine1Func) P((IN CONST compEnv *env,INOUT ExprPtr last,
					  INOUT ExprPtr *errlastp));
typedef ExprPtr (*compilePromoteFunc) P((IN CONST compEnv *env,INOUT ExprPtr last));

struct compilesimpleof {
  compileNeutralFunc neutral;
  compileCombineFunc combine;
  compileCombine1Func combine1;
  compilePromoteFunc promote;
  ExprPtr dompats;
  ExprPtr domains;
  ExprPtr locals;
  ExprPtr pat;
  ExprPtr f,p;
  Boolean mapsweep;
};

typedef ExprPtr (*compileBodyFunc) P((IN CONST ExprPtr x,IN CONST ExprPtr y,
				      IN CHAR *data,
				      IN CONST compEnv *env,INOUT ExprPtr last,
				      OUT ExprPtr *nonelastp,INOUT ExprPtr *errlastp));


IMPORT ExprPtr zApply P((IN CONST INT code,IN CONST ExprPtr arg));
IMPORT ExprPtr zApplyBin P((IN CONST INT code,
			    IN CONST ExprPtr arg1,IN CONST ExprPtr arg2));
IMPORT ExprPtr zApplyTer P((IN CONST INT code,
			    IN CONST ExprPtr arg1,IN CONST ExprPtr arg2,
			    IN CONST ExprPtr arg3));
IMPORT ExprPtr buildGnrlArg P((IN COUNTER n,IN CONST ExprPtr *vec));
IMPORT ExprPtr tackAssign P((IN CONST ExprPtr expr,INOUT ExprPtr last));
IMPORT ExprPtr tackApply P((IN CONST INT code,IN CONST ExprPtr arg,INOUT ExprPtr last));
IMPORT ExprPtr tackApplyBin P((IN CONST INT code,IN CONST ExprPtr arg1,
			       IN CONST ExprPtr arg2,INOUT ExprPtr last));
IMPORT ExprPtr tackApplyTer P((IN CONST INT code,IN CONST ExprPtr arg1,
			       IN CONST ExprPtr arg2,IN CONST ExprPtr arg3,
			       INOUT ExprPtr last));
IMPORT ExprPtr compileConstant P((IN CONST ExprPtr e,INOUT ExprPtr last,
				  IN CONST Boolean boxedp));
IMPORT ExprPtr compileIf P((IN CONST ExprPtr test,INOUT ExprPtr last,OUT ExprPtr *elselastp));
IMPORT ExprPtr mergeContinuations P((IN ExprPtr res1,INOUT ExprPtr last1,INOUT ExprPtr last2));

struct translation {
  ExprPtr prgm; /* the unshared list whose cdr is the actual statement list */
  	/* its car is the list of input arguments. */
};

IMPORT void initTranslation P((IN COUNTER arity,OUT struct translation *trans));
IMPORT ExprPtr makeCode P((IN CONST struct translation *fun,
			   IN CONST ExprPtr res));
IMPORT ExprPtr transToCode P((IN CONST struct translation *fun,
			      IN CONST ExprPtr res,INOUT ExprPtr last));

#define Z_NIL 0
#define Z_UNBOXED_INT 1
#define Z_UNBOX_INT 2
#define Z_BOX_INT 3
#define Z_UNBOXED_REAL 4
#define Z_UNBOX_REAL 5
#define Z_BOX_REAL 6
#define Z_UNBOX_TUPLE_NTH 7
#define Z_FAKE_UNBOX_TUPLE_LEN 8
#define Z_BOX_COMPLEX 9
#define Z_UNBOXED_STRING 10
#define Z_UNBOX_STRING_LEN 11
#define Z_UNBOX_STRING_TEXT 12
#define Z_BOX_STRING 13
#define Z_EMPTY 14
#define Z_TUPLE_0 15
#define Z_RECORD_0 16
#define Z_BOX_TYPE_APPL 17
#define Z_BOX_REF 18
#define Z_BOX_TUPLE_1 19
#define Z_BOX_UNSHARED_TUPLE_1 20
#define Z_BOX_RECORD_1 21
#define Z_BOX_UNSHARED_RECORD_1 22
#define Z_CODE 23
#define Z_QUEUE_TO_LIST 24
#define Z_BOX_CLOSURE 25
#define Z_BOX_CONS 26
#define Z_BOX_UNSHARED_CONS 27
#define Z_BOX_TUPLE_2 28
#define Z_BOX_UNSHARED_TUPLE_2 29

#define Z_BOX_TUPLE_GNRL 32
#define Z_BOX_UNSHARED_TUPLE_GNRL 33
#define Z_BOX_RECORD_GNRL 34
#define Z_BOX_UNSHARED_RECORD_GNRL 35
#define Z_BOX_SINGLETON 36
#define Z_OVERWRITE 37
#define Z_ENEXT 38
#define Z_VAR 39
#define Z_CONSENV 40
#define Z_RAISE 41
#define Z_ASSERT_EQUAL 42
#define Z_MAPPUT 43
#define Z_LET 44
#define Z_EQUAL 45
#define Z_CAR 46
#define Z_CDR 47
#define Z_UNBOX_REF 48
#define Z_HAS_TAG 49
#define Z_GET_VECTOR_LEN 50
#define Z_TYPE_MATCH 51
#define Z_MAP_CARD_DEC 52
#define Z_INT_LEQ 53
#define Z_GET_SHAPE 54
#define Z_MAPGET 55
#define Z_GET_MAPGET_Y 56
#define Z_GET_MAPGET_FOUND 57
#define Z_TUPLE_SUB_NTH 58
#define Z_MAPPUTBEHIND 59
#define Z_MAPREMOVE 60
#define Z_BOX_QUEUE 61
#define Z_UNBOXED_FALSE 62
#define Z_UNBOXED_TRUE 63
#define Z_BOX_1_QUEUE 64
#define Z_REF_SET 65
#define Z_BOX_NUMTYPE 66
#define Z_BOX_TYPEFX 67
#define Z_APPLY 68
#define Z_UNBOX_CLOSURE_FN 69
#define Z_UNBOX_CLOSURE_ENV 70
#define Z_APPEND 71
#define Z_MAPDOMRESTRTO 72
#define Z_MAPDOMRESTRBY 73
#define Z_START_LISTSWEEP 74
#define Z_INC_LISTSWEEP 75
#define Z_UNBOX_LISTSWEEP 76
#define Z_START_MAPSWEEP 77
#define Z_INC_MAPSWEEP 78
#define Z_UNBOX_MAPSWEEP 79
#define Z_START_SUBMAPSWEEP 80
#define Z_INC_SUBMAPSWEEP 81
#define Z_UNBOX_SUBMAPSWEEP 82
#define Z_BOX_LOCAL_REF 83
#define Z_UNBOX_LOCAL_REF 84
#define Z_LOCAL_REF_SET 85
#define Z_REVERSE1 86
#define Z_FAKE_MAPGET 87
#define Z_CONCAT_QUEUE 88
#define Z_BOX_SHARED_EXCEPTION 89
#define Z_BOX_UNSHARED_EXCEPTION 90
#define Z_FAKE_NOT_MAPGET 91
#define Z_BOX_BOOL 92
#define Z_UNBOX_BOOL 93
#define Z_START_HANDLE 94
#define Z_END_HANDLE 95
#define Z_C_GLOBAL 96
#define Z_RMU1 97
#define Z_UNSHARED_RMU1 98
#define Z_RMUX 99
#define Z_UNSHARED_RMUX 100
#define Z_MAKE_PROMISE 101
#define Z_BOX_PACK 102
#define Z_FAKE_MAPGET_FOUND 103
#define Z_MAPINTER 104
#define Z_CMPSTR 105
#define Z_CALL 106
#define Z_SET_C_GLOBAL 107
#define Z_C_CONST 108
#define Z_SET_C_CONST 109
#define Z_FAKEENV 110
#define Z_NOT 111
#define Z_GETTHEEXPR 112
