/* Support for mathematical functions.
   Copyright (C) 1992-2000 Jean Goubault-Larrecq and Bull S.A.
   Copyright (C) 2000-2005 Jean Goubault-Larrecq
   and LSV, CNRS UMR 8643 & INRIA Futurs projet Secsi & ENS Cachan.

   This file is part of HimML.

   HimML is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   HimML is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with HimML; see the file COPYING.  If not, write to
   the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "portable.h"
#include "ariproto.h"
#include <math.h>
#if OS==MAC
#ifndef THINK_C
#include <Types.h>
#endif
#endif
#ifdef HAS_PROTOTYPES
#include "smthutil.h"
#endif

#define __SEG__ SpecFuncUtils

EXTENDED
  min_float = 2.225073858E-308,
  min_extended = 1E-4919,
  double_eps = 3.552713679E-15,
  extend_eps = 5.421010863E-20,
  double_eps2 = 5.960464477E-8,
  extend_eps2 = 2.328306436E-10,
  double_2eps = 1.262177448E-29,
  extend_2eps = 2.938735877E-39,
  sqr2pi = 2.506628274631000295,
  sqrpi = 1.77245385090551602729,
  onepi = 3.14159265358979323846,
  twopi = 6.28318530717958647692,
  pi_over_two = 1.57079632679489661923,
  pi_over_four = 0.78539816339744830966,
  two_over_pi = 0.6366197723675813426,
  onee = 2.71828182845904523536,
  logsqr2pi = 0.91893853320467274178,
  logpi = 1.1447298858494001741434273,
  logtwo = 0.6931471805599453094172321,
  euler_gamma = 0.577215664901532860606512,
  zero = 0.0,
  fone = 1.0,
  fmone = -1.0,
  ftwo = 2.0,
  fhalf = 0.5,
  onethird = 0.333333333333333333333,
  onesixth = 0.1666666666666666666667,
  sqrt3 = 1.7320508075688772935274463;

Boolean rotten(x)
EXTENDED *x;
{
  short c;
	
  c = classifyX(x);
  return(c<C_ZERONUM && c>-C_ZERONUM);
}

void zmul(x1x,x1y,x2x,x2y,xx,xy)
EXTENDED *x1x,*x1y,*x2x,*x2y,*xx,*xy;
{
  EXTENDED aux;

  aux = (*x1x)*(*x2x)-(*x1y)*(*x2y);
  *xy = (*x1x)*(*x2y)+(*x2x)*(*x1y);
  *xx = aux;
}

void zdiv(x1x,x1y,x2x,x2y,xx,xy)
EXTENDED *x1x,*x1y,*x2x,*x2y,*xx,*xy;
{
  EXTENDED aux,r;

  r = (*x2x)*(*x2x)+(*x2y)*(*x2y);
  if (SMALLX(r))
    *xx = *xy = fone/zero;
  else
    {
      aux = ((*x1x)*(*x2x)+(*x1y)*(*x2y))/r;
      *xy = ((*x2x)*(*x1y)-(*x1x)*(*x2y))/r;
      *xx = aux;
    }
}

void zsqrt(x,y,xx,yy)
EXTENDED *x,*y,*xx,*yy;
{
  EXTENDED rr,theta;

  if (SMALLX(*y))
    {
      rr = sqrt(fabs(*x));
      if (classifyX(x)<0)	{ *xx = zero; *yy = rr; }
      else { *xx = rr; *yy = zero; }
    }
  else if (SMALLX(*x))
    {
      rr = sqrt(ldexp(fabs(*y),-1));
      if (classifyX(y)>0)	{ *xx = *yy = rr; }
      else { *xx = rr; *yy = -rr; }
    }
  else
    {
      rr = sqrt(sqrt((*x)*(*x)+(*y)*(*y)));
      theta = ldexp(atan2(*y,*x),-1);
      *xx = rr*cos(theta);
      *yy = rr*sin(theta);
    }
}

void zcubicroot(x,y,xx,yy)
EXTENDED *x,*y,*xx,*yy;
{
  EXTENDED rr,theta;
  
  if (SMALLX(*y))
    {
      *yy = zero;
      theta = fabs(*x);
      if (SMALLX(theta)) *xx = zero;
      else
	{
	  *xx = exp(onethird*log(theta));
	  if (classifyX(x)<0)	*xx = -(*xx);
	}
    }
  else if (SMALLX(*x))
    {
      theta = fabs(*y);
      *xx = exp(onethird*log(theta));
      *yy = ldexp(*xx,-1);
      *xx = sqrt3*(*yy);
      if (classifyX(y)<0)	*yy = -*yy;
    }
  else
    {
      rr = exp(onesixth*log((*x)*(*x)+(*y)*(*y)));
      theta = onethird*atan2(*y,*x);
      *xx = rr*cos(theta);
      *yy = rr*sin(theta);
    }
}

void zcos(x,y,xx,yy) /* cos(x+iy) = cos x.cos(iy) - sin x.sin(iy) = cos x.ch y - i.sin x.sh y */
EXTENDED *x,*y,*xx,*yy;
{
  *xx = cos(*x)*cosh(*y);
  *yy = -sin(*x)*sinh(*y);
}

void zsin(x,y,xx,yy) /* sin(x+iy) = sin x.cos(iy) + cos x.sin(iy) = sin x.ch y + i.cos x.sh y */
EXTENDED *x,*y,*xx,*yy;
{
  *xx = sin(*x)*cosh(*y);
  *yy = cos(*x)*sinh(*y);
}

void zch(x,y,xx,yy) /* ch(x+iy) = ch x.ch(iy) + sh x.sh(iy) = ch x.cos y + i.sh x.sin y */
EXTENDED *x,*y,*xx,*yy;
{
  *xx = cosh(*x)*cos(*y);
  *yy = sinh(*x)*sin(*y);
}

void zsh(x,y,xx,yy) /* sh(x+iy) = sh x.ch(iy) + ch x.sh(iy) = sh x.cos y + i.ch x.sin y */
EXTENDED *x,*y,*xx,*yy;
{
  *xx = sinh(*x)*cos(*y);
  *yy = cosh(*x)*sin(*y);
}

void rlog1(x,res) /* compute log(1+x) */
EXTENDED *x,*res;
{
  EXTENDED a;
  short n;
  
  a = fabs(*x);
  if (a>=fone)   *res = log(fone+(*x));
  else if (a<extend_eps)  *res = *x;
  else if (a<extend_eps2) *res = (*x)*(fone-ldexp(*x,-1));
  else /* log(1+x) = .5*log(1+2x+x^2) */
    {
      n = 0;
      a = *x;
      do {
	a = a*(a+ftwo);
	n--;
      } while  (fabs(a)<fone);
      *res = ldexp(log(fone+a),n);
    }
}

void zlog(x,y,xx,yy)
EXTENDED *x,*y,*xx,*yy;
{
  if (SMALLX(*y))
    {
      if (classifyX(x)>=0)   *yy = zero;
      else *yy = onepi;
      *xx = log(fabs(*x));
    }
  else if (SMALLX(*x))
    {
      if (classifyX(y)>0)	{ *xx = log(*y); *yy = pi_over_two; }
      else { *xx = log(-*y); *yy = -pi_over_two; }
    }
  else
    {
      *xx = ldexp(log((*x)*(*x)+(*y)*(*y)),-1);
      *yy = atan2(*y,*x);
    }
}

void zlog1(xx,xy,resx,resy) /* compute log(1+x) */
EXTENDED *xx,*xy,*resx,*resy;
{
  EXTENDED a,zx,zy;
  short n;
	
  if (SMALLX(*xy))
    {
      *resy = (classifyX(xx)>=0)?zero:onepi;
      zx = fabs(*xx);
      rlog1(&zx,resx);
    }
  else
    {
      a = (*xx)*(*xx)+(*xy)*(*xy);
      if (a>=fone)
	{
	  zx = fone+(*xx);
	  zlog(&zx,xy,resx,resy);
	}
      else if (a<extend_2eps)
	{
	  *resx = *xx;
	  *resy = *xy;
	}
      else if (a<extend_eps)
	{
	  zx = fone-ldexp(*xx,-1);
	  zy = -ldexp(*xy,-1); /* z = 1-x/2 */
	  *resx = (*xx)*zx-(*xy)*zy;
	  *resy = (*xy)*zx+(*xx)*zy;
	}
      else /* log(1+x) = .5*log(1+2x+x^2) */
	{
	  n = 0;
	  zx = *xx;
	  zy = *xy;
	  do {
	    a = zx+ftwo;
	    *resx = a*zx-zy*zy;
	    zy = (a+zx)*zy;
	    zx = *resx; /* z = z*(z+2) */
	    n--;
	  } while  (zx*zx+zy*zy<fone);
	  zx = fone+zx;
	  zlog(&zx,&zy,resx,resy);
	  *resx = ldexp(*resx,n);
	  *resy = ldexp(*resy,n);
	}
    }
}

void zth(x,y,xx,yy) /* th(x+iy) = (th(x)+th(iy))/(1+th(x).th(iy))
 = (th(x)+i.tan(y))/(1+i.th(x).tan(y)) */
EXTENDED *x,*y,*xx,*yy;
{
  EXTENDED nx,ny,rr,dy;
  
  nx = tanh(*x);
  ny = tan(*y);
  /* dx = fone; */
  dy = nx*ny;
  rr = fone+dy*dy;
  *xx = (nx+ny*dy)/rr;
  *yy = (ny-nx*dy)/rr;
}

void zath(x,y,xx,yy) /* ath(z = x+iy) = 1/2.log((1+z)/(1-z)) */
EXTENDED *x,*y,*xx,*yy;
{
  EXTENDED numx,numy,denx,deny,rx,ry;
  EXTENDED d;
  
  numx = ldexp(*x,1);
  numy = ldexp(*y,1);
  denx = fone - *x;
  deny = -*y;
  d = denx * denx +  deny * deny;
  rx = (numx * denx +  numy * deny) / d;
  ry = (numy * denx -  numx * deny) / d;
  zlog1(&rx,&ry,xx,yy);
  *xx = ldexp(*xx,-1);
  *yy = ldexp(*yy,-1);
}

void zsinpi(zx,zy,resx,resy)
EXTENDED *zx,*zy,*resx,*resy;
{
  EXTENDED x,y;
	
  x = -onepi*(*zx);
  y = -onepi*(*zy);
  zsin(&x,&y,resx,resy);
}

void ratan2(y,x,r)
EXTENDED *y,*x,*r;
{
  if (SMALLX(*x))
    {
      if (classifyX(y)>=0)   *r = pi_over_two;
      else *r = -pi_over_two;
    }
  else *r = atan2(*y,*x);
}

void zasin(x,y,xx,yy) /* asin(z) = -i.log(i.z + =(1-z^2)) */
EXTENDED *x,*y,*xx,*yy;
{
  Boolean neg;
  EXTENDED rx,ry,vx,vy,vvx,vvy;
  
  rx = *x;
  ry = *y;
  neg = (ry>extend_eps || (classifyX(&ry)>=0 && classifyX(&rx)<0));
  if (neg)   { rx = -rx; ry = -ry; }
  vvx = fone-rx*rx + ry*ry;
  vvy = -ldexp(rx*ry,1);
  zsqrt(&vvx,&vvy,&vx,&vy);
  vvx = -ry + vx;
  vvy = rx + vy;
  ratan2(&vvy,&vvx,xx);
  *yy = -ldexp(log(vvx*vvx+vvy*vvy),-1);
  if (neg)   { *xx = -*xx; *yy = -*yy; }
}

void zash(x,y,xx,yy) /* ash(z) = log(z + =(1+z^2)) */
EXTENDED *x,*y,*xx,*yy;
{
  Boolean neg;
  EXTENDED rx,ry,vx,vy,vvx,vvy;
  
  rx = *x;
  ry = *y;
  neg = (rx<zero);
  if (neg)   { rx = -rx; ry = -ry; }
  vvx = fone+rx*rx - ry*ry;
  vvy = ldexp(rx*ry,1);
  zsqrt(&vvx,&vvy,&vx,&vy);
  vvx = rx + vx;
  vvy = ry + vy;
  *xx = ldexp(log(vvx*vvx+vvy*vvy),-1);
  ratan2(&vvy,&vvx,yy);
  if (neg)   { *xx = -*xx; *yy = -*yy; }
}

void zacos(x,y,xx,yy) /* acos(z) = -i.log(z + =(z^2-1)) */ /* asin(z) = -i.log(i.z + =(1-z^2)) */
EXTENDED *x,*y,*xx,*yy;
{
  EXTENDED rx,ry,vx,vy,vvx,vvy;
  Boolean neg;
  
  rx = *x;
  ry = *y;
  vvx = rx*rx - ry*ry - fone;
  vvy = ldexp(rx*ry,1);
  zsqrt(&vvx,&vvy,&vx,&vy);
  neg = (rx<-extend_eps || (classifyX(&rx)<=0 && classifyX(&ry)<0));
  if (neg)
    {
      vvx = rx - vx;
      vvy = ry - vy;
    }
  else
    {
      vvx = rx + vx;
      vvy = ry + vy;
    }
  ratan2(&vvy,&vvx,xx);
  *yy = -ldexp(log(vvx*vvx+vvy*vvy),-1);
  if (neg)   { *xx = -*xx; *yy = -*yy; }
}

void zach(x,y,xx,yy) /* ach(z) = log(z + =(z^2-1)) */
EXTENDED *x,*y,*xx,*yy;
{
  EXTENDED rx,ry,vx,vy,vvx,vvy;
  
  rx = *x;
  ry = *y;
  vvx = rx*rx - ry*ry - fone;
  vvy = ldexp(rx*ry,1);
  zsqrt(&vvx,&vvy,&vx,&vy);
  vvx = rx + vx;
  vvy = ry + vy;
  *xx = ldexp(log(vvx*vvx+vvy*vvy),-1);
  ratan2(&vvy,&vvx,yy);
}

void zexp(x,y,xx,yy)
EXTENDED *x,*y,*xx,*yy;
{
  EXTENDED r;
  
  r = exp(*x);
  *xx = r*cos(*y);
  *yy = r*sin(*y);
}

void ztan(x,y,xx,yy) /* tan(x+iy) = (tan(x)+tan(iy))/(1-tan(x).tan(iy))
 = (tan(x)+i.th(y))/(1-i.tan(x).th(y)) */
EXTENDED *x,*y,*xx,*yy;
{
  EXTENDED nx,ny,rr,dy;
  
  if (SMALLX(*y))
    { *yy=zero; *xx=tan(*x); }
  else
    {
      nx = tan(*x);
      ny = tanh(*y);
      /* dx = fone; */
      dy = -nx*ny;
      rr = fone+dy*dy;
      *xx = (nx+ny*dy)/rr;
      *yy = (ny-nx*dy)/rr;
    }
}

void zatan(x,y,xx,yy) /* atan(z = x+iy) = i/2.log((i+z)/(i-z)) */
EXTENDED *x,*y,*xx,*yy;
{
  EXTENDED numx,numy,denx,deny,rx,ry;
  EXTENDED d;
  
  numx = *x;
  numy = fone + *y;
  denx = -*x;
  deny = fone-*y;
  d = denx * denx +  deny * deny;
  rx = (numx * denx +  numy * deny) / d;
  ry = (numy * denx -  numx * deny) / d;
  ratan2(&ry,&rx,xx);
  *xx = -ldexp(*xx,-1);
  *yy = ldexp(log(rx*rx+ry*ry),-2);
}

void rexp1(x,res) /* exp(x)-1 */
EXTENDED *x,*res;
{
  EXTENDED oldsum,del,kk;
  
  if (fabs(*x)<logtwo)
    {
      kk = ftwo;
      *res = del = (*x);
      do {
	oldsum = *res;
	del = del*(*x)/kk;
	*res = (*res)+del;
	kk = kk+fone;
      } while  ((*res)!=oldsum);
    }
  else *res = exp(*x)-fone;
}

void zexp1(xx,xy,resx,resy) /* exp(x)-1 */
EXTENDED *xx,*xy,*resx,*resy;
{
  EXTENDED zy,oldsumx,oldsumy,delx,dely,kk;
	
  zy = fmod(*xy,twopi);
  if (zy>onepi)  zy = zy-twopi;
  if (SMALLX(zy))
    {
      *resy = zero;
      rexp1(xx,resx);
    }
  else if (fabs(*xx)<logtwo && fabs(zy)<logtwo)
    {
      kk = ftwo;
      *resx = delx = (*xx);
      *resy = dely = zy;
      do {
	oldsumx = *resx;
	oldsumy = *resy;
	zmul(&delx,&dely,xx,&zy,&delx,&dely);
	delx = delx/kk;
	dely = dely/kk;
	*resx = (*resx)+delx;
	*resy = (*resy)+dely;
	kk = kk+fone;
      } while  ((*resx)!=oldsumx && (*resy)!=oldsumy);
    }
  else
    {
      zexp(xx,xy,resx,resy);
      *resx = (*resx)-fone;
    }
}

void rgd(x,res) /* gudermaniann */
EXTENDED *x,*res;
{
  /* gd x = 2.arctg exp x-=/2 ; gd is odd; for x small, gd x = x
     gd x = 2.arctg((exp x-1)/(exp x+1)) */
  Boolean pos;
  EXTENDED z,x1;
  
  pos = (classifyX(x)>0);
  z = pos?-(*x):(*x);
  rexp1(&z,&x1);
  *res = ldexp(atan(x1/(x1+ftwo)),1);
  if (pos)   *res = -(*res);
}

void zgd(xx,xy,resx,resy) /* gudermaniann */
EXTENDED *xx,*xy,*resx,*resy;
{
  /* gd x = 2.arctg exp x-=/2 ; gd is odd; for x small, gd x = x
     gd x = 2.arctg((exp x-1)/(exp x+1)) */
  Boolean pos;
  EXTENDED zx,zy,x1x,x1y,auxx,auxy;
  
  pos = (classifyX(xx)>0);
  if (pos)
    {
      zx = -(*xx);
      zy = -(*xy);
    }
  else
    {
      zx = (*xx);
      zy = (*xy);
    }
  zexp1(&zx,&zy,&x1x,&x1y);
  auxx = x1x+ftwo;
  auxy = x1y;
  zdiv(&x1x,&x1y,&auxx,&auxy,&auxx,&auxy);
  zatan(&auxx,&auxy,resx,resy);
  *resx = ldexp(*resx,1);
  *resy = ldexp(*resy,1);
  if (pos)
    {
      *resx = -(*resx);
      *resy = -(*resy);
    }
}

void zinvgd(xx,xy,resx,resy)
EXTENDED *xx,*xy,*resx,*resy;
{
  /* gd^-1(x) = log tg(=/4+z/2) = log((1+tg(z/2))/(1-tg(z/2))) */
  EXTENDED tx,ty,numx,numy,denx,deny,auxx,auxy;
  
  auxx = ldexp(*xx,-1);
  auxy = ldexp(*xy,-1);
  ztan(&auxx,&auxy,&tx,&ty);
  zlog1(&tx,&ty,&numx,&numy);
  auxx = -auxx;
  auxy = -auxy;
  zlog1(&tx,&ty,&denx,&deny);
  *resx = numx-denx;
  *resy = numy-deny;
}

void zpow(xx,xy,yx,yy,resx,resy)
EXTENDED *xx,*xy,*yx,*yy,*resx,*resy;
{
  EXTENDED auxx,auxy,dumx,dumy;

  zlog(xx,xy,&auxx,&auxy);
  zmul(&auxx,&auxy,yx,yy,&dumx,&dumy);
  zexp(&dumx,&dumy,resx,resy);
}
